import {
  type Component,
  type JSX,
  splitProps,
  type Ref,
  createRenderEffect,
  onCleanup,
  children,
  createEffect,
  createSignal,
  onMount,
} from "solid-js";
import { Dynamic, type DynamicProps } from "solid-js/web";
import MasonryLayout from "masonry-layout";
import { createElementSize } from "@solid-primitives/resize-observer";
import "./Masonry.css";

type MasonryContainer =
  | Exclude<keyof JSX.IntrinsicElements, keyof JSX.SVGElementTags>
  | Component<{
      ref?: Ref<Element>;
      class?: string;
      children?: JSX.Element;
    }>;

type ElementOf<T extends MasonryContainer> =
  T extends Exclude<keyof JSX.IntrinsicElements, keyof JSX.SVGElementTags>
    ? JSX.IntrinsicElements[T] extends { ref?: Ref<infer E> }
      ? E
      : never
    : T extends Component<{ ref?: Ref<infer E> }>
      ? E
      : never;

function forwardRef<T>(value: T, ref?: Ref<T>) {
  if (!ref) return;
  (ref as (value: T) => void)(value);
}

function createMasonry(element: Element, options: () => MasonryLayout.Options) {
  const layout = new MasonryLayout(element, {
    initLayout: false,
  });

  onCleanup(() => layout.destroy?.());

  const size = createElementSize(element);

  createRenderEffect(() => {
    const opts = options();
    layout.option?.(opts);
  });

  createRenderEffect(() => {
    const width = size.width; // only tracking width
    layout.layout?.();
  });

  if (import.meta.hot) {
    import.meta.hot.on("vite:afterUpdate", () => {
      layout.layout?.();
    });
  }

  return layout;
}

const supportsCSSMasonryLayout = /* @__PURE__ */ CSS.supports(
  "grid-template-rows",
  "masonry",
);

console.debug("supports css masonry layout", supportsCSSMasonryLayout);

const useNativeImpl = import.meta.env.VITE_PLATFROM_MASONRY_ALWAYS_COMPAT
  ? false
  : supportsCSSMasonryLayout;

if (import.meta.env.VITE_PLATFROM_MASONRY_ALWAYS_COMPAT) {
  console.warn(
    "Masonry is in compat mode because VITE_PLATFORM_MASONRY_ALWAYS_COMPAT is enabled",
  );
}

function MasonryCompat<T extends MasonryContainer>(
  oprops: DynamicProps<T> & { class?: string },
) {
  const [props, rest] = splitProps(oprops, ["ref", "children", "class"]);

  const childrenComponents = children(() => props.children);

  return (
    <Dynamic
      ref={(element: ElementOf<T>) => {
        forwardRef(element, props.ref as Ref<typeof element> | undefined);

        const [columnGap, setColumnGap] = createSignal<number>();

        const layout = createMasonry(element, () => {
          return {
            gutter: columnGap(),
          };
        });

        createEffect(() => {
          const computedStyle = window.getComputedStyle(element);

          const rowGap = computedStyle.rowGap;
          if (element instanceof HTMLElement) {
            element.style.setProperty("--Masonry-row-gap", rowGap);
          }

          const colGap = computedStyle.columnGap;
          if (colGap) {
            setColumnGap(Number(colGap.slice(0, colGap.length - 2)));
          }
        });

        createRenderEffect(() => {
          childrenComponents(); // just tracks
          setTimeout(() => {
            layout.reloadItems?.();
            layout.layout?.();
          }, 0);
        });
      }}
      class={`Masonry CompatMasonry ${props.class || ""}`}
      {...rest}
      children={childrenComponents}
    />
  );
}

function MasonryNative<T extends MasonryContainer>(
  oprops: DynamicProps<T> & { class?: string },
) {
  const [props, rest] = splitProps(oprops, ["class"]);
  return (
    <Dynamic class={`Masonry NativeMasonry ${props.class || ""}`} {...rest} />
  );
}

/**
 * Masonry Layout Container.
 *
 * **Native if possible** This component uses css masonry layout
 * and fallback to masonry-layout if not supported. The children
 * must have specified width and height.
 *
 * **Children Changes** As the children changed, reflow will be triggered,
 * and there is might be a blink (or transition) for user. If it's not your
 * intention, don't remove/add the direct children. Instead wraps them under
 * containers and set the width and height on the container.
 *
 * **CSS compatibility** This component compatible to "gap" "row-gap"
 * "column-gap" property. But they are read only once after the element mounted.
 */
export default useNativeImpl ? MasonryNative : MasonryCompat;
